module.exports = (function() {

  var path = require("path");
  var fs = require("fs");  
  var util = require("util");
  var Promise = require("bluebird");
  var _ = require("lodash");
  var moment = require("moment");
  var TDXapi = require("nqm-api-tdx");
  var reformat = require("./reformat");
  var convertor = require("./event-store-convertor");
  var checkNewLine = require("./check-newline");
  var _output;
  var _tdxApi = null;
  var _config = null;
  var _sourceFilePreparationProgress;   // Estimate source file preparation will take 25% of total time.
  var _progress = 0;
  var _progressStep = 3; 

  var databot = function(instanceId, authToken, inputData, output, control) {
    // Combine all inputs into config object.
    _config = _.extend({}, control.packageParams, inputData);
    _config.instanceId = instanceId;
    _config.accessToken = authToken;

    _output = output;
    _output.progress(0);

    checkArguments();  

    initialiseTDXApi()
    .then(getSourceFile)
    .then(createDataset)
    .then(getTargetDataset)
    .then(prepareSourceFiles)
    .then(runImport)
    .then(function() {
      return setDatasetImporting(false);
    })
    .then(function() {
      _output.debug("finished");
      _output.result({ok: true, datasetId: _config.targetDataset});
      _output.progress(100);        
    })
    .catch(function(err) {
      _output.debug("uncaught exception [%s]", err.message);
      _output.debug(err.stack);
      process.exit(1);
    });
  };

  var abort = function() {
    var msg = util.format.apply(util, arguments);
    _output.debug("aborting - " + msg);
    _output.error(msg);
    process.exit(1);
  };

  var checkArguments = function() {
    // Allow defaults for some parameters.
    _config.readHost = _config.readHost || "localhost:27017";
    _config.readDb = _config.readDb || "nqmQuery";
    _config.esHost = _config.esHost || "localhost:27017";
    _config.esDb = _config.esDb || "nqmES";
    // _config.commandHost = _config.commandHost || "https://cmd.nqminds.com";
    // _config.queryHost = _config.queryHost || "https://q.nqminds.com";
    _config.copyNameIndex = 0;
    
    /*
     * Check system (i.e. non-task) arguments.
     * These should be configured when the process is installed.
     */
    if (!_config.readDb) {
      abort("no read db connection string");
    }
    if (!_config.esDb) {
      abort("no event store db connection string");
    }
    if (!_config.commandHost) {
      abort("no command host given");
    }
    if (!_config.queryHost) {
      abort("no query host given");
    }
    if (!_config.resourceStorePath) {
      abort("no resource store path");
    }

    /*
     * Check task-instance arguments.
     * These are specified by the user when the task is started.
     */
    if (!_config.sourceResource) {
      abort("no source resource given");
    }
    if (!_config.accessToken) {
      abort("no accessToken given");
    }
    if (!_config.targetDataset) {
      if (!_config.targetSchema) {
        abort("no targetSchema given");
      }
      if (!_config.targetFolder) {
        abort("no targetFolder given");
      }
      if (!_config.targetName) {
        _config.targetName = _config.targetSchema + " import " + moment().format("Do MMM YYYY");
      }
    }
  };

  var initialiseTDXApi = function() {
    _tdxApi = new TDXapi(_config);
    Promise.promisifyAll(_tdxApi);
    updateProgress();
    return Promise.resolve();
  };

  var getSourceFile = function() {
    _config.sourceFile = path.join(_config.resourceStorePath, _config.sourceResource + ".json");
    try {
      // Check file is accessible.
      fs.accessSync(_config.sourceFile, fs.F_OK);
      updateProgress();
    } catch (e) {
      _output.debug("can't access source file " + _config.sourceFile + " error is: " + e.message);
      abort("sourceResource input '%s' is invalid - it must be a TDX resource of type rawFile/json with data uploaded.",_config.sourceResource);
    }
  };

  var createDataset = function() {
    var datasetPromise;

    if (!_config.targetDataset) {
      var targetName = _config.copyNameIndex > 0 ? _config.targetName + " copy " + _config.copyNameIndex : _config.targetName;
      // Create the dataset with 'importing' flag set so that the data projection does not start until the import is complete.
      datasetPromise = _tdxApi.createDatasetAsync({ name: targetName, basedOnSchema: _config.targetSchema, parentId: _config.targetFolder, importing: true })
        .then(function(createResult) {
          _output.debug("created dataset id: %j", createResult);
          _config.targetDataset = createResult.response.id;

          updateProgress();
        })
        .catch(function(err) {
          // TODO - add error codes to the entire TDX!
          if (err.message.indexOf("resource already exists with name:") >= 0) {
            _output.debug("name clash - bumping number and trying again");
            _config.copyNameIndex++;
            return createDataset();
          } else {
            abort("unable to create target dataset '%s' using schema '%s' in folder '%s' [%s]", targetName, _config.targetSchema, _config.targetFolder, err.message);
          }
        });
    } else {
      var truncatePromise;

      if (_config.truncateTarget) {
        _output.debug("truncating target dataset");
        truncatePromise = truncateDataset();
      } else {
        _output.debug("NOT truncating target dataset");
        truncatePromise = Promise.resolve();
      }
      // Switch off data projection
      _output.debug("importing to existing dataset - switching off projection");
      updateProgress();
      datasetPromise = truncatePromise
        .then(function() { 
          updateProgress();
          return setDatasetImporting(true); 
        });
    }
    
    return datasetPromise
      .then(function() {
        // Delay here to give the projection time to catch up.
        return new Promise(function(resolve, reject) {
          setTimeout(function() {
            updateProgress();
            resolve(_config.targetDataset);
          }, 10000);
        });        
      });
  };

  var getTargetDataset = function() {
    return _tdxApi.queryAsync("datasets/" + _config.targetDataset, null, null, null)
      .then(function(dataset) {
        _output.debug("retrieved dataset %j", dataset);
        _config.dataset = dataset;

        if (!_config.targetName) {
          _config.targetName = dataset.name;
        }

        updateProgress();
        return Promise.resolve(dataset);
      })
      .catch(function(err) {
        abort("failed to get target dataset '%s' [%s]", _config.targetDataset, err.message);
      });
  };

  var setDatasetImporting = function(importing) {
    if (importing) {
      _output.debug("importing flag set - stopping data projection");
    } else {
      _output.debug("importing flag cleared - starting data projection");
    }
    // Clear the dataset 'importing' flag so that the data projection can start.
    return _tdxApi.setDatasetImportFlagAsync(_config.targetDataset, importing)
      .then(function(result) {
        _output.debug("setDatasetImporting to '%s' result: %j", importing, result);
        return Promise.resolve();
      })
      .catch(function(err) {
        abort("failed to start data projection for dataset '%s' [%s]", _config.targetDataset, err.message);
      });
  };

  var truncateDataset = function() {
    // Truncate dataset data
    return _tdxApi.truncateDatasetAsync(_config.targetDataset, false)
      .then(function(result) {
        _output.debug("truncateDataset result: %j", result);
        return Promise.resolve();
      })
      .catch(function(err) {
        abort("failed to truncate dataset '%s' [%s]", _config.targetDataset, err.message);
      });
  };

  var doImport = function(importIdx, dbHost, dbName, collectionName, sourceFile) {
    var spawn = require("child_process").spawn;

    var totalProgress = (100-_sourceFilePreparationProgress)/2;

    _output.debug("starting import of %s into %s/%s/%s", sourceFile, dbHost, dbName, collectionName);
    return new Promise(function(resolve, reject) {
      var importArgs = ["--host",dbHost,"--db",dbName,"--collection", collectionName,"--file", sourceFile];
      var importChild = spawn("mongoimport", importArgs);
      
      // Intercept mongoimport STDOUT - attempt to parse the progress string. 
      importChild.stdout.on("data", function(stdout) {
        stdout = stdout.toString();
        _output.debug("import - %s", stdout.trim());
        var progressLabel = "%";
        var progressIdx = stdout.indexOf(progressLabel); 
        if (progressIdx >= 0) {
          var progress = parseInt(stdout.substr(progressIdx - 2,3).trim());
          
          // Adjust progress by _sourceFilePreparationProgress and the fact there are two imports to run.
          var adjustedProgress = _sourceFilePreparationProgress + totalProgress*importIdx + totalProgress*(progress/100);
          _output.progress(adjustedProgress);
        }      
      });

      importChild.on("close", function(code) {
        _output.debug("exited with code %d", code);
        if (code) {
          abort("Import failed with exit code %d", code);
        } else {
          // Finised with zero exit code - update final progress.
          _output.progress(_sourceFilePreparationProgress + totalProgress*(importIdx+1));
          resolve();
        }
      });

      importChild.on("error", function(err) {
        _output.error("importer error: " + err.message);
      });
    });
  };

  var runImport = function() {
    _output.debug("runImport");

    _sourceFilePreparationProgress = _progress;

    var dataset = _config.dataset;

    // First import the event store collection.
    return doImport(0, _config.esHost, _config.esDb, dataset.store + ".events", _config.eventSourceFile)
      .then(function() {
        // Then import the read-side collection.
        return doImport(1, _config.readHost, _config.readDb, dataset.store, _config.readSourceFile);
      })
      .catch(function(err) {
        abort("unexpected error during core import: %s", err.message);
      });
  };

  var prepareSourceFiles = function() {
    _output.debug("preparing source files for import");

    return checkNewLine(_config.sourceFile, _output)
      .then(function() {
        updateProgress();
        _config.eventSourceFile = _config.sourceFile + ".es.json";
        if (_config.sourceDataPath) {
          // A data path is given, so assume the files need re-formatting from proper JSON 
          // with sourceDataPath pointing to the array to import, rather than new-line separated JSON documents
          // as required by mongoimport.
          _output.debug("!!!!!!");
          _output.debug("sourceDataPath %s => assuming array-based OLD FORMAT", _config.sourceDataPath);
          _output.debug("!!!!!!");
          _config.readSourceFile = _config.sourceFile + ".read.json";
          return reformat(_config.sourceDataPath + ".*", _config.sourceFile, _config.readSourceFile, _config.eventSourceFile, _output)
            .then(function() {
              updateProgress();
            })
            .catch(function(err) {
              _output.debug(err.stack);          
              abort("failure during source file re-formatting: %s", err.message);
            });
        } else {
          // No data path => just need to create the event-store version of the source file.
          _output.debug("!!!!!!");
          _output.debug("no sourceDataPath => assuming newline-delimited NEW FORMAT");
          _output.debug("!!!!!!");
          _config.readSourceFile = _config.sourceFile;
          return convertor(_config.sourceFile, _config.eventSourceFile, _output)
            .then(function() {
              updateProgress();
            })
            .catch(function(err) {
              _output.debug(err.stack);          
              abort("failure during conversion to event-store format: %s", err.message);
            });
        }
      });
  };

  var updateProgress = function() {
    _progress += _progressStep;
    _output.progress(_progress);
  };

  return databot;
}());
